/* $Id: String.cc,v 1.2 2002/04/28 19:29:00 sdevaux Exp $
 *
 * NaJA
 * Tabs = 4 chars
 * copyright (C) 2002 Sebastien Devaux
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 *              Sebastien Devaux <sebastien.devaux@laposte.net>
 */ 
#include <math.h>
#include "lang/String.hh"

PACKAGE_BEGIN(najaLang)

char * String::null="";
//----------------------------------------------
// Default constructor
String::String() : Object() {
	MSTART("String::String()")
	value=String::null;
	len=0;
	MEND
}
//----------------------------------------------
// count chars to the next \0
int String::strlen(const char * str) {
	MSTART("String::strlen(\""<<str<<"\")")
	int count=0;
	for (const char * c=str;*c!='\0';c++,count++);
	MRETURN(count)
}	
//----------------------------------------------
// native c string copy
void String::strcpy(char * dest,const char * src,int lmax) {
	MSTART("String::strcpy(\""<<dest<<"\",\""<<src<<"\","<<lmax<<")")
	int i;
	for (i=0;i<lmax && src[i]!='\0';i++)
		dest[i]=src[i];
	dest[i]='\0';
	MEND
}
//----------------------------------------------
// native c string compare
int String::strcmp(const char * s1,const char * s2,int lmax) {
	MSTART("String::strcmp(\""<<s1<<"\",\""<<s2<<"\","<<lmax<<")")
	int diff=0;
	for (int i=0;i<lmax && diff==0;i++)
		diff=s1[i]-s2[i];
	MRETURN(diff)
}
//----------------------------------------------
// native c string compare ignoring case
int String::strcasecmp(const char * s1,const char * s2,int lmax) {
	MSTART("String::strcasecmp(\""<<s1<<"\",\""<<s2<<"\","<<lmax<<")")
	int diff=0;
	for (int i=0;i<lmax && (diff==0 || diff==32);i++)
		diff=s1[i]-s2[i];
	if (diff==32) return 0;
	MRETURN(diff)
}
//----------------------------------------------
// new String concatenated from two other strings
String::String(const char * str1,const char * str2) {
	MSTART("String::String(\""<<str1<<"\",\""<<str2<<"\")")
	len=strlen(str1)+strlen(str2);
	value=new char[len+1];
	int i;
	for (i=0;str1[i]!='\0';i++)
		value[i]=str1[i];
	for (int j=0;str2[j]!='\0';i++,j++)
		value[i]=str2[j];
	value[i]='\0';	
	MEND
} 
//----------------------------------------------
// free localy allocated memory
String::~String() {
	if (value!=NULL && value!=String::null) {
		delete value;
	}
}	
//----------------------------------------------
// new String replacing chars from another one
String::String(NAJA<String> str, char oldChar, char newChar) {
	MSTART("String::String(\""<<str<<"\",'"<<oldChar<<"','"<<newChar<<"')")
	len=str->len;
	value=new char[len+1];
	for (int i=0;i<=len;i++) {
		if (str->value[i]==oldChar)
			value[i]=newChar;
		else
			value[i]=str->value[i];
	}
	MEND
}
//----------------------------------------------
// new String
String::String(const char * str) {
	MSTART("String::String("<<str<<")")
	len=strlen(str);
	value=new char[len+1];
	strcpy(value,str,len);
	value[len]='\0';
	MEND
}
//----------------------------------------------
// new String from another string part
String::String(const char * str,int offset,
				int count) {
	MSTART("String::String("<<str<<","<<offset<<","<<count<<")")
	len=count;
	value=new char[len+1];
	strcpy(value,&(str[offset]),len);
	value[len]='\0';
	MEND
}
//----------------------------------------------
// get one char from the String
char String::charAt(int index) {
	MSTART("String::charAt("<<index<<")")
	MRETURN(value[index])
}
char String::operator[](int index) {
	return value[index];
}
//----------------------------------------------
// this String's length
int String::length() {
	return len;
}	
//----------------------------------------------
// copy part of this String to a native char array
void String::getChars(int srcBegin,int srcEnd,char * dst,int dstBegin) {
	MSTART("String::getChars("<<srcBegin<<","<<srcEnd<<","<<dst
			<<","<<dstBegin<<")")
	int srci;
	int dsti;
	for (srci=srcBegin,dsti=dstBegin;srci<srcEnd;srci++,dsti++)
		dst[dsti]=value[srci];
	MEND
}
//----------------------------------------------
// equality 
boolean String::equals(NAJA<String> anotherString) {
	MSTART("String::equals(\""<<anotherString<<"\")")
	MRETURN(strcmp(value,anotherString->value,len+1)==0)
}
//----------------------------------------------
// equality ignoring case
boolean String::equalsIgnoreCase(NAJA<String> anotherString) {
	MSTART("String::equalsIgnoreCase(\""<<anotherString<<"\")")
	MRETURN(strcasecmp(value,anotherString->value,len+1)==0)
}
//----------------------------------------------
// String compare
int String::compareTo(NAJA<String> anotherString) {
	MSTART("String::compareTo(\""<<anotherString<<"\")")
	MRETURN(strcmp(value,anotherString->value,len+1))
}
//----------------------------------------------
// String compare ignoring case
int String::compareToIgnoreCase(NAJA<String> anotherString) {
	MSTART("String::compareToIgnoreCase(\""<<anotherString<<"\")")
	MRETURN(strcasecmp(value,anotherString->value,len+1))
}
//----------------------------------------------
// part equality
boolean String::regionMatches(int toffset,NAJA<String> other,
											int ooffset,int l) {
	MSTART("String::regionMatches("<<toffset<<",\""<<other<<"\","
			<<ooffset<<","<<l<<")")
	MRETURN(strcmp(&(value[toffset]),&(other->value[ooffset]),l)==0)
}
//----------------------------------------------
// part equality 
boolean String::regionMatch(boolean ignoreCase,int toffset,
                            NAJA<String> other,int ooffset,int l) {
	MSTART("String::regionMatch("<<ignoreCase<<","<<toffset<<",\""
			<<other<<"\","<<ooffset<<","<<l<<")")
	if (ignoreCase) { 
		MRETURN(strcasecmp(&(value[toffset]),&(other->value[ooffset]),l)==0)
	}
	else {
		MRETURN(strcmp(&(value[toffset]),&(other->value[ooffset]),l)==0)
	}
}
//----------------------------------------------
// check string prefix
boolean String::startsWith(NAJA<String> prefix,int offset) {
	MSTART("String::startsWith(\""<<prefix<<"\","<<offset<<")")
	MRETURN(strcmp(&(value[offset]),prefix->value,prefix->len)==0)
}
//----------------------------------------------
// check string prefix
boolean String::startsWith(NAJA<String> prefix) {
	MSTART("String::startsWith(\""<<prefix<<"\")")
	return (strcmp(value,prefix->value,prefix->len)==0);
}
//----------------------------------------------
// check string suffix
boolean String::endsWith(NAJA<String> suffix) {
	MSTART("String::endsWith(\""<<suffix<<"\")")
	MRETURN(strcmp(&(value[len-suffix->len]),suffix->value,suffix->len)==0)
}
//----------------------------------------------
// hash code
int String::hashCode() {
	MSTART("String::hashCode")
	int i;
	int p;
	int res=0;
	for (i=0,p=len-1;p>=0;i++,p++)
		res+=((int)value[i])*(int)pow(31,p);
	MRETURN(res)
}
//----------------------------------------------
// forward char search
int String::indexOf(int ch,int fromIndex) {
	MSTART("String::indexOf('"<<ch<<"',"<<fromIndex<<")")
	if (fromIndex < 0) {
		fromIndex=0;
	} else if (fromIndex>=len) {
		MRETURN(-1)
	}
	for (int i=fromIndex;i<len;i++) {
		if (ch==value[i]) {
			MRETURN(i)
		}
	}
	MRETURN(-1)
}
//----------------------------------------------
// backward char search
int String::lastIndexOf(int ch,int fromIndex) {
	MSTART("String::lastIndexOf('"<<ch<<"',"<<fromIndex<<")")
	if (fromIndex <0) {
		MRETURN(-1)
	} else if (fromIndex>=len) {
		fromIndex=len-1;
	}
	for (int i=fromIndex;i>=0;i--) {
		if (ch==value[i])
			MRETURN(i)
	}
	MRETURN(-1)
}
//----------------------------------------------
// forward substring search
int String::indexOf(NAJA<String> str, int fromIndex) {
	MSTART("String::indexOf(\""<<str<<"\","<<fromIndex<<")")
	char * first=&(str->value[0]);
	int pos=fromIndex-1;
	while ((pos=indexOf(*first,++pos))>=0) {
		if (startsWith(str,pos))
			MRETURN(pos)
	}
	MRETURN(-1)
}	
//----------------------------------------------
// backward substring search
int String::lastIndexOf(NAJA<String> str, int fromIndex) {
	MSTART("String::lastIndexOf(\""<<str<<"\","<<fromIndex<<")")
	char* first=&(str->value[0]);
	int pos=fromIndex+1;
	while ((pos=lastIndexOf(*first,--pos))>=0) {
		if (startsWith(str,pos))
			MRETURN(pos)
	}
	MRETURN(-1)
}
//----------------------------------------------
// extract substring
NAJA<String> String::substring(int beginIndex, int endIndex) {
	MSTART("String::substring("<<beginIndex<<","<<endIndex<<")")
	if (endIndex>len) endIndex=len;
	MRETURN(new String(value,beginIndex,endIndex-beginIndex))
}
//----------------------------------------------
// append a string 
NAJA<String> String::concat(NAJA<String> str) {
	MSTART("String::concat(\""<<str<<"\")")
	MRETURN(new String(value,str->value))
}
//----------------------------------------------
// char replacement
NAJA<String> String::replace(char oldChar,char newChar) {
	MSTART("String::replace('"<<oldChar<<"','"<<newChar<<")")
	MRETURN(new String(this,oldChar,newChar))
}
//----------------------------------------------
// delete space chars at both ends of string
NAJA<String> String::trim() {
	MSTART("String::trim()")
	int end=len;
	int start=0;
	while ((start<end) && (value[start]<=' '))
		start++;
	while ((start<end) && (value[end]<=' '))
		end--;
	if (start>0 || end<len) {
		MRETURN(new String(value,start,end-start))
	}
	else {
		MRETURN(this)
	}
}
//----------------------------------------------
// toString() is natural !
NAJA<String> String::toString() {
	MSTART("String::toString()")
	MRETURN(this)
}
//----------------------------------------------
// get native c string
const char* String::c_str() {
	MSTART("String::c_str()")
	MRETURN(value)
}
//----------------------------------------------
// copy this string
NAJA<Object> String::clone() {
	MSTART("String::clone()")
	MRETURN(new String(value))
}

//----------------------------------------------
// operator + overload : concat String with String
// or native c string
NAJA<String> operator+(NAJA<String> str1,NAJA<String> str2) {
	return new String(str1->value,str2->value);
}
NAJA<String> operator+(NAJA<String> str1,const char * str2) {
	return new String(str1->value,str2);
}
NAJA<String> operator+(const char * str1,NAJA<String> str2) {
	return new String(str1,str2->value);
}

PACKAGE_END
